import requests
from python_helper import Constant as c
from python_helper import ReflectionHelper, ObjectHelper, log, Function

from python_framework.api.src.constant import HttpClientConstant, LogConstant
from python_framework.api.src.domain import HttpDomain
from python_framework.api.src.util import FlaskUtil
from python_framework.api.src.util import Serializer
from python_framework.api.src.converter.static import ConverterStatic
from python_framework.api.src.enumeration.HttpStatus import HttpStatus
from python_framework.api.src.service.flask import FlaskManager
from python_framework.api.src.service.openapi import OpenApiManager
from python_framework.api.src.util import ClientUtil
from python_framework.api.src.util.ClientUtil import HttpClientEvent, ManualHttpClientEvent


@Function
def HttpClient(url=c.BLANK, headers=None, timeout=HttpClientConstant.DEFAULT_TIMEOUT, logRequest=False, logResponse=False) :
    clientUrl = url
    clientHeaders = ConverterStatic.getValueOrDefault(headers, dict())
    clientTimeout = timeout
    clientLogRequest = logRequest
    clientLogResponse = logResponse
    def Wrapper(OuterClass, *args, **kwargs):
        log.wrapper(HttpClient,f'''wrapping {OuterClass.__name__}''')
        class InnerClass(OuterClass):
            url = clientUrl
            headers = clientHeaders
            timeout = clientTimeout
            logRequest = clientLogRequest
            logResponse = clientLogResponse
            def __init__(self,*args, **kwargs):
                log.wrapper(OuterClass,f'in {InnerClass.__name__}.__init__(*{args},**{kwargs})')
                apiInstance = FlaskManager.getApi()
                OuterClass.__init__(self,*args, **kwargs)
                self.globals = apiInstance.globals
            def options(self, *args, **kwargs):
                raise HttpClientEvent(HttpDomain.Verb.OPTIONS, *args, **kwargs)
            def get(self, *args, **kwargs):
                raise HttpClientEvent(HttpDomain.Verb.GET, *args, **kwargs)
            def post(self, *args, **kwargs):
                raise HttpClientEvent(HttpDomain.Verb.POST, *args, **kwargs)
            def put(self, *args, **kwargs):
                raise HttpClientEvent(HttpDomain.Verb.PUT, *args, **kwargs)
            def patch(self, *args, **kwargs):
                raise HttpClientEvent(HttpDomain.Verb.PATCH, *args, **kwargs)
            def delete(self, *args, **kwargs):
                raise HttpClientEvent(HttpDomain.Verb.DELETE, *args, **kwargs)
        ReflectionHelper.overrideSignatures(InnerClass, OuterClass)
        return InnerClass
    return Wrapper


class ClientMethodConfig:
    def __init__(self,
        url = None,
        headers = None,
        requestHeaderClass = None,
        requestParamClass = None,
        requestClass = None,
        responseClass = None,
        returnOnlyBody = None,
        timeout = None,
        propagateAuthorization = None,
        propagateApiKey = None,
        propagateSession = None,
        produces = None,
        consumes = None,
        logRequest = None,
        logResponse = None
    ):
        self.url = url
        self.headers = headers
        self.requestHeaderClass = requestHeaderClass
        self.requestParamClass = requestParamClass
        self.requestClass = requestClass
        self.responseClass = responseClass
        self.returnOnlyBody = returnOnlyBody
        self.timeout = timeout
        self.propagateAuthorization = propagateAuthorization
        self.propagateApiKey = propagateApiKey
        self.propagateSession = propagateSession
        self.produces = produces
        self.consumes = consumes
        self.logRequest = logRequest
        self.logResponse = logResponse


@Function
def HttpClientMethod(
    url = c.BLANK,
    headers = None,
    requestHeaderClass = None,
    requestParamClass = None,
    requestClass = None,
    responseClass = None,
    returnOnlyBody = True,
    timeout = None,
    propagateAuthorization = False,
    propagateApiKey = False,
    propagateSession = False,
    consumes = OpenApiManager.DEFAULT_CONTENT_TYPE,
    produces = OpenApiManager.DEFAULT_CONTENT_TYPE,
    logRequest = False,
    logResponse = False,
    debugIt = False
):
    clientMethodConfig = ClientMethodConfig(
        url = url,
        headers = headers,
        requestHeaderClass = requestHeaderClass,
        requestParamClass = requestParamClass,
        requestClass = requestClass,
        responseClass = responseClass,
        returnOnlyBody = returnOnlyBody,
        timeout = timeout,
        propagateAuthorization = propagateAuthorization,
        propagateApiKey = propagateApiKey,
        propagateSession = propagateSession,
        produces = produces,
        consumes = consumes,
        logRequest = logRequest,
        logResponse = logResponse,
    )
    def innerMethodWrapper(resourceInstanceMethod,*args, **kwargs) :

        def options(
            resourceInstance,
            body = None,
            additionalUrl = None,
            params = None,
            headers = None,
            timeout = None,
            logRequest = False,
            **kwargs
        ):
            verb = HttpDomain.Verb.OPTIONS
            url, params, headers, body, timeout, logRequest = ClientUtil.parseParameters(
                resourceInstance,
                clientMethodConfig,
                additionalUrl,
                params,
                headers,
                body,
                timeout,
                logRequest
            )
            doLogRequest(verb, url, body, params, headers, logRequest, kwargs)
            clientResponse = requests.options(
                url,
                params = params,
                headers = headers,
                json = body,
                timeout = timeout,
                **kwargs
            )
            return clientResponse

        def get(
            resourceInstance,
            body = None,
            additionalUrl = None,
            params = None,
            headers = None,
            timeout = None,
            logRequest = False,
            **kwargs
        ):
            verb = HttpDomain.Verb.GET
            url, params, headers, body, timeout, logRequest = ClientUtil.parseParameters(
                resourceInstance,
                clientMethodConfig,
                additionalUrl,
                params,
                headers,
                body,
                timeout,
                logRequest
            )
            doLogRequest(verb, url, body, params, headers, logRequest, kwargs)
            clientResponse = requests.get(
                url,
                params = params,
                headers = headers,
                json = body,
                timeout = timeout,
                **kwargs
            )
            return clientResponse

        def post(
            resourceInstance,
            body = None,
            additionalUrl = None,
            headers = None,
            params = None,
            timeout = None,
            logRequest = False,
            **kwargs
        ):
            verb = HttpDomain.Verb.POST
            url, params, headers, body, timeout, logRequest = ClientUtil.parseParameters(
                resourceInstance,
                clientMethodConfig,
                additionalUrl,
                params,
                headers,
                body,
                timeout,
                logRequest
            )
            doLogRequest(verb, url, body, params, headers, logRequest, kwargs)
            clientResponse = requests.post(
                url,
                params = params,
                headers = headers,
                json = body,
                timeout = timeout,
                **kwargs
            )
            return clientResponse

        def put(
            resourceInstance,
            body = None,
            additionalUrl = None,
            headers = None,
            params = None,
            timeout = None,
            logRequest = False,
            **kwargs
        ):
            verb = HttpDomain.Verb.PUT
            url, params, headers, body, timeout, logRequest = ClientUtil.parseParameters(
                resourceInstance,
                clientMethodConfig,
                additionalUrl,
                params,
                headers,
                body,
                timeout,
                logRequest
            )
            doLogRequest(verb, url, body, params, headers, logRequest, kwargs)
            clientResponse = requests.put(
                url,
                params = params,
                headers = headers,
                json = body,
                timeout = timeout,
                **kwargs
            )
            return clientResponse

        def patch(
            resourceInstance,
            body = None,
            additionalUrl = None,
            headers = None,
            params = None,
            timeout = None,
            logRequest = False,
            **kwargs
        ):
            verb = HttpDomain.Verb.PATCH
            url, params, headers, body, timeout, logRequest = ClientUtil.parseParameters(
                resourceInstance,
                clientMethodConfig,
                additionalUrl,
                params,
                headers,
                body,
                timeout,
                logRequest
            )
            doLogRequest(verb, url, body, params, headers, logRequest, kwargs)
            clientResponse = requests.patch(
                url,
                params = params,
                headers = headers,
                json = body,
                timeout = timeout,
                **kwargs
            )
            return clientResponse

        def delete(
            resourceInstance,
            body = None,
            additionalUrl = None,
            headers = None,
            params = None,
            timeout = None,
            logRequest = False,
            **kwargs
        ):
            verb = HttpDomain.Verb.DELETE
            url, params, headers, body, timeout, logRequest = ClientUtil.parseParameters(
                resourceInstance,
                clientMethodConfig,
                additionalUrl,
                params,
                headers,
                body,
                timeout,
                logRequest
            )
            doLogRequest(verb, url, body, params, headers, logRequest, kwargs)
            clientResponse = requests.delete(
                url,
                params = params,
                headers = headers,
                json = body,
                timeout = timeout,
                **kwargs
            )
            return clientResponse

        def doLogRequest(verb, url, body, params, headers, logRequest, requestKwargs):
            log.info(resourceInstanceMethod, f'{LogConstant.CLIENT_SPACE}{verb}{c.SPACE_DASH_SPACE}{url}')
            if logRequest:
                parsetRequestKwargs = {} if ObjectHelper.isEmpty(requestKwargs) else {'requestKwargs': {**requestKwargs}}
                log.prettyJson(
                    resourceInstanceMethod,
                    LogConstant.CLIENT_REQUEST,
                    {
                        'headers': ConverterStatic.getValueOrDefault(headers, dict()),
                        'query': ConverterStatic.getValueOrDefault(params, dict()),
                        'body': ConverterStatic.getValueOrDefault(body, dict()),
                        **parsetRequestKwargs
                    },
                    condition = True,
                    logLevel = log.INFO
                )

        HTTP_CLIENT_RESOLVERS_MAP = {
            HttpDomain.Verb.OPTIONS : options,
            HttpDomain.Verb.GET : get,
            HttpDomain.Verb.POST : post,
            HttpDomain.Verb.PUT : put,
            HttpDomain.Verb.PATCH : patch,
            HttpDomain.Verb.DELETE : delete
        }

        log.wrapper(HttpClientMethod,f'''wrapping {resourceInstanceMethod.__name__}''')
        def innerResourceInstanceMethod(*args, **kwargs):
            f'''(*args, {FlaskUtil.KW_HEADERS}={{}}, {FlaskUtil.KW_PARAMETERS}={{}}, **kwargs)'''
            resourceInstance = args[0]
            clientResponse = None
            completeResponse = None
            try :
                FlaskManager.validateKwargs(
                    kwargs,
                    resourceInstance,
                    resourceInstanceMethod,
                    requestHeaderClass,
                    requestParamClass
                )
                FlaskManager.validateArgs(args, requestClass, resourceInstanceMethod)
                clientResponse = None
                httpClientEvent = ClientUtil.getHttpClientEvent(resourceInstanceMethod, *args, **kwargs)
                if isinstance(httpClientEvent, ManualHttpClientEvent):
                    completeResponse = httpClientEvent.completeResponse
                elif isinstance(httpClientEvent, HttpClientEvent):
                    try :
                        clientResponse = HTTP_CLIENT_RESOLVERS_MAP.get(
                            httpClientEvent.verb,
                            ClientUtil.raiseHttpClientEventNotFoundException
                        )(
                            resourceInstance,
                            *httpClientEvent.args,
                            **httpClientEvent.kwargs
                        )
                    except Exception as exception:
                        ClientUtil.raiseException(clientResponse, exception)
                    ClientUtil.raiseExceptionIfNeeded(clientResponse)
                    completeResponse = ClientUtil.getCompleteResponse(clientResponse, responseClass, produces)
                    FlaskManager.validateCompleteResponse(responseClass, completeResponse)
                else:
                    raise Exception('Unknown http client event')
            except Exception as exception:
                log.log(innerResourceInstanceMethod, 'Failure at client method execution', exception=exception, muteStackTrace=True)
                FlaskManager.raiseAndPersistGlobalException(exception, resourceInstance, resourceInstanceMethod, context=HttpDomain.CLIENT_CONTEXT)
            clientResponseStatus = completeResponse[-1]
            clientResponseHeaders = completeResponse[1]
            clientResponseBody = completeResponse[0] if ObjectHelper.isNotNone(completeResponse[0]) else {'message' : HttpStatus.map(clientResponseStatus).enumName}
            if resourceInstance.logResponse or logResponse :
                log.prettyJson(
                    resourceInstanceMethod,
                    LogConstant.CLIENT_RESPONSE,
                    {
                        'headers': clientResponseHeaders,
                        'body': Serializer.getObjectAsDictionary(clientResponseBody, muteLogs=not debugIt),
                        'status': clientResponseStatus
                    },
                    condition = True,
                    logLevel = log.INFO
                )
            if returnOnlyBody:
                return completeResponse[0]
            else:
                return completeResponse
        ReflectionHelper.overrideSignatures(innerResourceInstanceMethod, resourceInstanceMethod)
        innerResourceInstanceMethod.url = clientMethodConfig.url
        innerResourceInstanceMethod.headers = clientMethodConfig.headers
        innerResourceInstanceMethod.requestHeaderClass = clientMethodConfig.requestHeaderClass
        innerResourceInstanceMethod.requestParamClass = clientMethodConfig.requestParamClass
        innerResourceInstanceMethod.requestClass = clientMethodConfig.requestClass
        innerResourceInstanceMethod.responseClass = clientMethodConfig.responseClass
        innerResourceInstanceMethod.returnOnlyBody = clientMethodConfig.returnOnlyBody
        innerResourceInstanceMethod.timeout = clientMethodConfig.timeout
        innerResourceInstanceMethod.propagateAuthorization = clientMethodConfig.propagateAuthorization
        innerResourceInstanceMethod.propagateApiKey = clientMethodConfig.propagateApiKey
        innerResourceInstanceMethod.propagateSession = clientMethodConfig.propagateSession
        innerResourceInstanceMethod.produces = clientMethodConfig.produces
        innerResourceInstanceMethod.consumes = clientMethodConfig.consumes
        innerResourceInstanceMethod.logRequest = clientMethodConfig.logRequest
        innerResourceInstanceMethod.logResponse = clientMethodConfig.logResponse

        return innerResourceInstanceMethod
    return innerMethodWrapper



from python_helper import Constant as c

REQUEST = 'Request'
RESPONSE = 'Response'

CONTROLLER_SPACE = f'[CONTROLLER]{c.SPACE}'
CONTROLLER_REQUEST = f'{CONTROLLER_SPACE}{REQUEST}'
CONTROLLER_RESPONSE = f'{CONTROLLER_SPACE}{RESPONSE}'

CLIENT_SPACE = f'[CLIENT]{c.SPACE}'
CLIENT_REQUEST = f'{CLIENT_SPACE}{REQUEST}'
CLIENT_RESPONSE = f'{CLIENT_SPACE}{RESPONSE}'

LISTENER_SPACE = f'[LISTENER]{c.SPACE}'
LISTENER_REQUEST = f'{LISTENER_SPACE}{REQUEST}'
LISTENER_RESPONSE = f'{LISTENER_SPACE}{RESPONSE}'

EMITTER_SPACE = f'[EMITTER]{c.SPACE}'
EMITTER_REQUEST = f'{EMITTER_SPACE}{REQUEST}'
EMITTER_RESPONSE = f'{EMITTER_SPACE}{RESPONSE}'



import datetime
from python_helper import DateTimeHelper, ObjectHelper

def now():
    return datetime.datetime.utcnow()

def ofTimestamp(t):
    return datetime.datetime.utcfromtimestamp(t)
    
def plusMinutes(givenDateTime, minutes=None, deltaInMinutes=None) :
    if ObjectHelper.isNotNone(minutes) :
        deltaInMinutes = DateTimeHelper.timeDelta(seconds=minutes*60)
    return givenDateTime + deltaInMinutes



from python_framework.api.src.enumeration.HttpStatus import HttpStatus
from python_framework.api.src.service.flask.FlaskManager import Controller, ControllerMethod
import TestCallDto

@Controller(url = '/call', tag='Call', description='Call controller')
class TestCallController:

    @ControllerMethod(
        requestClass = TestCallDto.TestCallRequestDto,
        responseClass = TestCallDto.TestCallRequestDto
    )
    def post(self, dto):
        return dto, HttpStatus.OK

@Controller(url = '/call/batch', tag='Call', description='Call controller')
class TestCallBatchController:

    @ControllerMethod(
        requestClass = [[TestCallDto.TestCallRequestDto]],
        responseClass = [[TestCallDto.TestCallRequestDto]]
    )
    def post(self, dtoList):
        return dtoList, HttpStatus.OK



from python_framework.api.src.annotation.EnumAnnotation import Enum, EnumItem

@Enum()
class CallStatusEnumeration :
    INCOMMING = EnumItem()
    WASTED = EnumItem()

CallStatus = CallStatusEnumeration()



import importlib
importlib.invalidate_caches()
from python_helper import EnvironmentHelper, log, RandomHelper
EnvironmentHelper.update("URL_VARIANT", RandomHelper.integer(minimum=0, maximum=10000))
log.debug(log.debug, f'variant: {EnvironmentHelper.get("URL_VARIANT")}')

from python_helper import TestHelper
TestHelper.run(__file__)
# TestHelper.run(
#     __file__,
#     runOnly = [
#         'SessionManagerTest.sessionManager_worksProperly',
#         'ApiKeyManagerTest.apiKeyManager_worksProperly'
#     ]
# )
# TestHelper.run(
#     __file__,
#     # times=10,
#     runOnly = [
#         # 'TestApiTest.appRun_whenEnvironmentIsLocalFromLocalConfig_withSuccess',
#         # 'TestApiTest.pythonRun_securityManager',
#         # 'TestApiTest.pythonRun_apiKeyManager',
#         # 'TestApiTest.pythonRun_sessionManager',
#         'TestApiTest.testing_headersAndParams'
#         # ,
#         # 'TestApiTest.testing_Client'
#     ]
# )
# TestHelper.run(
#     __file__,
#     runOnly = [
#         'EnumAnnotationTest.Enum_comparing'
#     ]
# )
# TestHelper.run(
#     __file__,
#     runOnly = [
#         'GlobalExceptionAnnotationTest.encapsulateItWithGlobalException_withParameters_GlobalException'
#     ]
# )
# TestHelper.run(
#     __file__,
#     runOnly = [
#         'EnumAnnotationTest.enum_withSuccess',
#         'EnumAnnotationTest.otherEnum_withSuccess',
#         'EnumAnnotationTest.python_framework_status',
#         'EnumAnnotationTest.enumName',
#         'EnumAnnotationTest.enumName_badImplementation',
#         'EnumAnnotationTest.map_whenArgIsNone',
#         'EnumAnnotationTest.Enum_whenHaveMoreThanOneInnerValue',
#         'EnumAnnotationTest.Enum_dot_map',
#         'EnumAnnotationTest.Enum_str',
#         'EnumAnnotationTest.Enum_strInOutput'
#     ]
# )
# TestHelper.run(
#     __file__,
#     runOnly = [
#         'SerializerTest.convertFromObjectToObject_whenTargetClassIsList',
#         'SerializerTest.isModelTest',
#         'SerializerTest.isJsonifyable',
#         'SerializerTest.jsonifyIt',
#         'SerializerTest.fromDtoToModel',
#         'SerializerTest.fromModelToDto',
#         'SerializerTest.convertFromJsonToObject_whenThereAreEnums',
#         'SerializerTest.convertFromObjectToObject_weirdIdList',
#         'SerializerTest.convertFromJsonToObject_nativeClassAtributeList'
#     ]
# )
# TestHelper.run(
#     __file__,
#     runOnly = [
#         'SerializerTest.fromModelToDto',
#         'SerializerTest.fromDtoToModel'
#     ]
# )
# TestHelper.run(
#     __file__,
#     runOnly = [
#         'SerializerTest.convertFromJsonToObject_whenThereAreEnums'
#     ]
# )

log.debug(log.debug, f'variant: {EnvironmentHelper.get("URL_VARIANT")}')



from python_helper import log
log.debug(log, 'imported')

from python_framework import HttpStatus
log.debug(HttpStatus, 'imported')

from python_framework.api.src.model import ErrorLog
log.debug(ErrorLog, 'imported')

from python_framework.api.src.util import Serializer
log.debug(Serializer, 'imported')

from python_framework.api.src.service.ExceptionHandler import GlobalException
log.debug(GlobalException, 'imported')

from python_framework.api.src.service import SecurityManager
log.debug(SecurityManager, 'imported')

from python_framework.api.src.service import SqlAlchemyProxy
log.debug(SqlAlchemyProxy, 'imported')

from python_framework.api.src.service.openapi import OpenApiManager
log.debug(OpenApiManager, 'imported')

from python_framework.api.src.service.flask import FlaskManager
log.debug(FlaskManager, 'imported')

from python_framework.api.src.service.flask import ResourceManager
log.debug(ResourceManager, 'imported')

from python_framework.api.src.annotation import EnumAnnotation
log.debug(EnumAnnotation, 'imported')



from flask import Response, request
import flask_restful

from python_helper import Constant as c
from python_helper import ObjectHelper, log, Function
import globals

from python_framework.api.src.util import Serializer
from python_framework.api.src.enumeration.HttpStatus import HttpStatus


KEY_API_INSTANCE = 'apiInstance'
API_INSTANCE_HOLDER = {
    KEY_API_INSTANCE: None
}

KW_PARAMETERS = 'params'
KW_HEADERS = 'headers'

UNKNOWN_VERB = 'UNKNOWN'

CLIENT_RESPONSE_TEXT = '__clientResponseText__'
CLIENT_RESPONSE = '__clientResponse__'

Response = Response
request = request
Resource = flask_restful.Resource

@Function
def safellyGetRequestBody():
    requestBody = safellyGetJson()
    if ObjectHelper.isNone(requestBody):
        requestBody = safellyGetData()
    return requestBody if ObjectHelper.isNotNone(requestBody) else dict()

@Function
def safellyGetJson():
    jsonBody = None
    try :
        jsonBody = request.get_json(force=True)
    except Exception as exception:
        jsonBody = {}
        log.log(safellyGetJson, f'Not possible to get request body. Returning {jsonBody} by default', exception=exception)
    return jsonBody

@Function
def safellyGetFlaskResponseJson(response: Response):
    jsonBody = None
    try :
        jsonBody = response.get_json(force=True)
    except Exception as exception:
        jsonBody = {}
        log.log(safellyGetFlaskResponseJson, f'Not possible to get response body. Returning {jsonBody} by default', exception=exception)
    return jsonBody

@Function
def safellyGetResponseJson(response):
    jsonBody = None
    try :
        jsonBody = response.json()
    except Exception as exception:
        jsonBody = {
            CLIENT_RESPONSE_TEXT: response.text,
            CLIENT_RESPONSE: response
        }
        log.log(safellyGetResponseJson, f'Not possible to get response body. Returning {jsonBody} by default', exception=exception)
    return jsonBody

@Function
def safellyGetData():
    dataBody = None
    try :
        dataBody = request.get_json(force=True)
    except Exception as exception:
        dataBody = {}
        log.log(safellyGetData, f'Not possible to get data. Returning {dataBody} by default', exception=exception)
    return dataBody

@Function
def safellyGetUrl():
    url = None
    try :
        url = str(request.url)
    except Exception as exception:
        log.log(safellyGetUrl, 'Not possible to get url', exception=exception)
    return url

def safellyGetHost():
    host = None
    try :
        path = str(request.host)
    except Exception as exception:
        host = c.FOWARD_SLASH
        log.log(safellyGetUrl, f'Not possible to get host. Returning {host} by default', exception=exception)
    return host

def safellyGetPath():
    path = None
    try :
        path = str(request.path)
    except Exception as exception:
        path = c.FOWARD_SLASH
        log.log(safellyGetPath, f'Not possible to get path. Returning {path} by default', exception=exception)
    return path

@Function
def safellyGetVerb():
    verb = None
    try :
        verb = str(request.method)
    except Exception as exception:
        log.log(safellyGetVerb, 'Not possible to get verb', exception=exception)
    return verb

@Function
def safellyGetHeaders():
    headers = None
    try:
        headers = dict(request.headers)
    except Exception as exception:
        headers = {}
        log.log(safellyGetHeaders, f'Not possible to get request headers. Returning {headers} by default', exception=exception)
    return headers if ObjectHelper.isNotNone(headers) else dict()

@Function
def safellyGetResponseHeaders(response):
    headers = None
    try:
        headers = dict(response.headers)
    except Exception as exception:
        headers = {}
        log.log(safellyGetResponseHeaders, f'Not possible to get response headers. Returning {headers} by default', exception=exception)
    return headers if ObjectHelper.isNotNone(headers) else dict()

@Function
def safellyGetRequestHeadersFromResponse(response):
    headers = None
    try:
        headers = dict(response.request.headers)
    except Exception as exception:
        headers = {}
        log.log(safellyGetRequestHeadersFromResponse, f'Not possible to get request headers from response. Returning {headers} by default', exception=exception)
    return headers if ObjectHelper.isNotNone(headers) else dict()

@Function
def safellyGetArgs():
    args = None
    try:
        args = dict(request.args)
    except Exception as exception:
        args = {}
        log.log(safellyGetArgs, f'Not possible to get args. Returning {args} by default', exception=exception)
    return args if ObjectHelper.isNotNone(args) else dict()

@Function
def safellyGetRequestJsonFromResponse(response):
    requestBody = None
    try:
        requestBody = response.request.body
    except Exception as exception:
        requestBody = {}
        log.log(safellyGetRequestJsonFromResponse, f'Not possible to get request body from response. Returning {requestBody} by default', exception=exception)
    return requestBody if ObjectHelper.isNotNone(requestBody) else dict()

@Function
def safellyGetRequestUrlFromResponse(response):
    url = None
    try:
        url = response.url
    except Exception as exception:
        url = c.BLANK
        log.log(safellyGetRequestUrlFromResponse, f'Not possible to get request url from response. Returning {url} by default', exception=exception)
    return url if ObjectHelper.isNotNone(url) else c.BLANK

@Function
def safellyGetRequestVerbFromResponse(response):
    verb = None
    try:
        verb = response.request.method
    except Exception as exception:
        verb = UNKNOWN_VERB
        log.log(safellyGetRequestUrlFromResponse, f'Not possible to get request verb from response. Returning {verb} by default', exception=exception)
    return verb

@Function
def safellyGetResponseStatus(response):
    status = None
    try:
        status = response.status_code
    except Exception as exception:
        status = HttpStatus.INTERNAL_SERVER_ERROR
        log.log(safellyGetResponseStatus, f'Not possible to get response status. Returning {status} by default', exception=exception)
    try:
        status = HttpStatus.map(status if ObjectHelper.isNotNone(status) else HttpStatus.INTERNAL_SERVER_ERROR)
    except Exception as exception:
        status = HttpStatus.INTERNAL_SERVER_ERROR
        log.log(safellyGetResponseStatus, f'Not possible to parse response status. Returning {status} by default', exception=exception)
    return status

@Function
def buildHttpResponse(additionalResponseHeaders, controllerResponseBody, status, contentType):
    httpResponse = Response(Serializer.jsonifyIt(controllerResponseBody),  mimetype=contentType, status=status)
    for key, value in additionalResponseHeaders.items():
        httpResponse.headers[key] = value
    return httpResponse

@Function
def addToKwargs(key, givenClass, valuesAsDictionary, kwargs):
    if ObjectHelper.isNotEmpty(givenClass):
        toClass = givenClass if ObjectHelper.isNotList(givenClass) else givenClass[0]
        kwargs[key] = Serializer.convertFromJsonToObject({k:v for k,v in valuesAsDictionary.items()}, toClass)
    return valuesAsDictionary

@Function
def getGlobals():
    return globals.getGlobalsInstance()

@Function
def getApi():
    api = None
    try:
        api = getGlobals().api
    except Exception as exception:
        raise Exception(f'Failed to return api from "globals" instance. Cause: {str(exception)}')
    return api

def getNullableApi():
    api = None
    try :
        api = getApi()
    except Exception as exception:
        log.warning(getNullableApi, 'Not possible to get api', exception=exception)
    return api

@Function
def getClassName(instance):
    return instance.__class__.__name__

@Function
def getModuleName(instance):
    return instance.__class__.__module__

@Function
def getQualitativeName(instance):
    return instance.__class__.__qualname__

@Function
def isApiInstance(apiInstance):
    if ObjectHelper.isNone(apiInstance):
        return False
    apiClassName = flask_restful.Api.__name__
    moduleName = flask_restful.__name__
    return apiClassName == getClassName(apiInstance) and apiClassName == getQualitativeName(apiInstance) and moduleName == getModuleName(apiInstance)

def validateFlaskApi(apiInstance):
    if not isApiInstance(apiInstance):
        raise Exception(f'Invalid "flask_restful.Api" instance. {apiInstance} is not an Api instance')

def validateResourceInstance(resourceInstance):
    if ObjectHelper.isNone(resourceInstance):
        raise Exception(f'Resource cannot be None')

@Function
def retrieveApiInstance(apiInstance=None, arguments=None):
    if isApiInstance(apiInstance):
        return apiInstance
    if isApiInstance(API_INSTANCE_HOLDER.get(KEY_API_INSTANCE)):
        return API_INSTANCE_HOLDER.get(KEY_API_INSTANCE)
    if ObjectHelper.isNone(apiInstance) and ObjectHelper.isNotNone(arguments):
        try:
            apiInstance = arguments[0].globals.api
        except Exception as exception:
            log.warning(retrieveApiInstance, f'''Not possible to retrieve api instance by {arguments}. Going for another approach''', exception=exception)
    if not isApiInstance(apiInstance):
        log.warning(retrieveApiInstance, f'''Not possible to retrieve api instance. Going for a slower approach''')
        apiInstance = getNullableApi()
    if ObjectHelper.isNone(apiInstance):
        raise Exception('Not possible to retrieve api instance')
    API_INSTANCE_HOLDER[KEY_API_INSTANCE] = apiInstance
    return apiInstance



from python_helper import log, Test, SettingHelper, RandomHelper, ObjectHelper, TestHelper, ReflectionHelper, Constant
from python_framework import EncapsulateItWithGlobalException, GlobalException, ExceptionHandler, HttpStatus

LOG_HELPER_SETTINGS = {
    log.LOG : False,
    log.INFO : True,
    log.SUCCESS : True,
    log.SETTING : True,
    log.DEBUG : True,
    log.WARNING : True,
    log.WRAPPER : True,
    log.FAILURE : True,
    log.ERROR : True,
    log.TEST : False
}

FULL_LOG_HELPER_SETTINGS = {
    log.LOG : True,
    log.INFO : True,
    log.SUCCESS : True,
    log.SETTING : True,
    log.DEBUG : True,
    log.WARNING : True,
    log.WRAPPER : True,
    log.FAILURE : True,
    log.ERROR : True,
    log.TEST : False
}

SUCCESS = '__SUCCESS__'
FAILURE = '__FAILURE__'

RAISED_EXCEPTION = Exception(FAILURE)

@EncapsulateItWithGlobalException()
def externalFuncionDoesNotThrowsException():
    return SUCCESS

@EncapsulateItWithGlobalException()
def externalFuncionDoesThrowsException():
    raise RAISED_EXCEPTION

@Test(environmentVariables={
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **FULL_LOG_HELPER_SETTINGS
    }
)
def encapsulateItWithGlobalException_noParameters_unknownException() :
    #arrange
    @EncapsulateItWithGlobalException()
    def internalFuncionDoesNotThrowsException():
        return SUCCESS
    @EncapsulateItWithGlobalException()
    def internalFuncionDoesThrowsException():
        raise RAISED_EXCEPTION

    # #act
    externalSuccess = externalFuncionDoesNotThrowsException()
    internalSuccess = internalFuncionDoesNotThrowsException()
    externalFailure = TestHelper.getRaisedException(externalFuncionDoesThrowsException)
    internalFailure = TestHelper.getRaisedException(internalFuncionDoesThrowsException)
    # print(externalFailure.logResource)
    # print(externalFailure.logResourceMethod)
    # print(internalFailure.logResource)
    # print(internalFailure.logResourceMethod)

    #assert
    assert SUCCESS == externalSuccess
    assert SUCCESS == internalSuccess
    assert not RAISED_EXCEPTION == externalFailure, f'not {RAISED_EXCEPTION} == {externalFailure}: {not RAISED_EXCEPTION == externalFailure}'
    assert not RAISED_EXCEPTION == internalFailure, f'not {RAISED_EXCEPTION} == {internalFailure}: {not RAISED_EXCEPTION == internalFailure}'
    assert not RAISED_EXCEPTION == externalFailure, f'not {RAISED_EXCEPTION} == {externalFailure}: {not RAISED_EXCEPTION == externalFailure}'
    assert not RAISED_EXCEPTION == internalFailure, f'not {RAISED_EXCEPTION} == {internalFailure}: {not RAISED_EXCEPTION == internalFailure}'
    assert GlobalException == ReflectionHelper.getClass(externalFailure)
    assert GlobalException == ReflectionHelper.getClass(internalFailure)
    assert ExceptionHandler.DEFAULT_MESSAGE == externalFailure.message, f'{ExceptionHandler.DEFAULT_LOG_MESSAGE} == {externalFailure.message}: {ExceptionHandler.DEFAULT_LOG_MESSAGE == externalFailure.message}'
    assert ExceptionHandler.DEFAULT_MESSAGE == internalFailure.message
    assert HttpStatus.INTERNAL_SERVER_ERROR == externalFailure.status
    assert HttpStatus.INTERNAL_SERVER_ERROR == internalFailure.status
    assert FAILURE == externalFailure.logMessage
    assert FAILURE == internalFailure.logMessage
    assert ExceptionHandler.DEFAULT_LOG_RESOURCE == externalFailure.logResource, f'{ExceptionHandler.DEFAULT_LOG_RESOURCE} == {externalFailure.logResource}: {ExceptionHandler.DEFAULT_LOG_RESOURCE == externalFailure.logResource}'
    assert externalFuncionDoesThrowsException.__name__ == externalFailure.logResourceMethod.__name__, f'{externalFuncionDoesThrowsException} == {externalFailure.logResourceMethod}: {externalFuncionDoesThrowsException == externalFailure.logResourceMethod}'
    assert ExceptionHandler.DEFAULT_LOG_RESOURCE == internalFailure.logResource, f'{ExceptionHandler.DEFAULT_LOG_RESOURCE} == {internalFailure.logResource}: {ExceptionHandler.DEFAULT_LOG_RESOURCE == internalFailure.logResource}'
    assert type(internalFuncionDoesThrowsException) == type(internalFailure.logResourceMethod)


@Test(environmentVariables={
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **FULL_LOG_HELPER_SETTINGS
    }
)
def encapsulateItWithGlobalException_noParameters_GlobalException() :
    #arrange

    class MyClass:
        def myMethod(self):
            ...
    resource = MyClass()
    ERROR_MESSAGE = 'ERROR_MESSAGE'
    LOG_ERROR_MESSAGE = 'LOG_ERROR_MESSAGE'
    EXCEPTION_STATUS = HttpStatus.BAD_REQUEST
    simpleException = GlobalException(
        status = EXCEPTION_STATUS,
        message = ERROR_MESSAGE,
        logMessage = LOG_ERROR_MESSAGE,
        logResource = resource,
        logResourceMethod = resource.myMethod
    )
    @EncapsulateItWithGlobalException()
    def internalFuncionDoesThrowsException():
        raise simpleException

    # #act
    internalFailure = TestHelper.getRaisedException(internalFuncionDoesThrowsException)
    # print(internalFailure)
    # print(internalFailure.logResource)
    # print(internalFailure.logResourceMethod)

    #assert
    assert not RAISED_EXCEPTION == internalFailure, f'not {RAISED_EXCEPTION} == {internalFailure}: {not RAISED_EXCEPTION == internalFailure}'
    assert not RAISED_EXCEPTION == internalFailure, f'not {RAISED_EXCEPTION} == {internalFailure}: {not RAISED_EXCEPTION == internalFailure}'
    assert GlobalException == ReflectionHelper.getClass(internalFailure)
    assert ERROR_MESSAGE == internalFailure.message, f'"{ERROR_MESSAGE}" and "{internalFailure.message}" should be equals'
    assert EXCEPTION_STATUS == internalFailure.status
    assert LOG_ERROR_MESSAGE == internalFailure.logMessage, f'"{LOG_ERROR_MESSAGE} == {internalFailure.logMessage}": {LOG_ERROR_MESSAGE == internalFailure.logMessage}'
    assert resource == internalFailure.logResource, f'"{resource} == {internalFailure.logResource}": {resource == internalFailure.logResource}'
    assert resource.myMethod == internalFailure.logResourceMethod, f'"{resource.myMethod} == {internalFailure.logResourceMethod}": {resource.myMethod == internalFailure.logResourceMethod}'


@Test(environmentVariables={
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **FULL_LOG_HELPER_SETTINGS
    }
)
def encapsulateItWithGlobalException_withParameters_GlobalException() :
    #arrange

    class MyClass:
        def myMethod(self):
            ...
    resource = MyClass()
    ERROR_MESSAGE = 'ERROR_MESSAGE'
    LOG_ERROR_MESSAGE = 'LOG_ERROR_MESSAGE'
    EXCEPTION_STATUS = HttpStatus.BAD_REQUEST
    simpleException = Exception(ERROR_MESSAGE)
    PERSONALIZED_MESSAGE = 'PERSONALIZED_MESSAGE'
    PERSONALIZED_STATUS = HttpStatus.UNAUTHORIZED
    @EncapsulateItWithGlobalException(message=PERSONALIZED_MESSAGE, status=PERSONALIZED_STATUS)
    def internalFuncionDoesThrowsException():
        raise simpleException

    #act
    internalFailure = TestHelper.getRaisedException(internalFuncionDoesThrowsException)
    # print(internalFailure)
    # print(internalFailure.logResource)
    # print(internalFailure.logResourceMethod)

    #assert
    assert not RAISED_EXCEPTION == internalFailure, f'not {RAISED_EXCEPTION} == {internalFailure}: {not RAISED_EXCEPTION == internalFailure}'
    assert not RAISED_EXCEPTION == internalFailure, f'not {RAISED_EXCEPTION} == {internalFailure}: {not RAISED_EXCEPTION == internalFailure}'
    assert GlobalException == ReflectionHelper.getClass(internalFailure)
    assert PERSONALIZED_MESSAGE == internalFailure.message, f'{PERSONALIZED_MESSAGE} == {internalFailure.message}: {PERSONALIZED_MESSAGE == internalFailure.message}'
    assert PERSONALIZED_STATUS == internalFailure.status
    assert ERROR_MESSAGE == internalFailure.logMessage, f'{ERROR_MESSAGE} == {internalFailure.logMessage}: {ERROR_MESSAGE == internalFailure.logMessage}'
    assert ExceptionHandler.DEFAULT_LOG_RESOURCE == internalFailure.logResource
    assert type(internalFuncionDoesThrowsException) == type(internalFailure.logResourceMethod)



from python_helper import EnvironmentHelper
from python_framework.api.src.enumeration.HttpStatus import HttpStatus
from python_framework.api.src.service import SecurityManager
from python_framework.api.src.service.flask.FlaskManager import Controller, ControllerMethod
from python_framework.api.src.util import UtcDateTimeUtil


VALID_TOKEN_MINUTES_DURATION = 30


@Controller(url=f'/test/{EnvironmentHelper.get("URL_VARIANT")}/security-manager', tag='SecurityManagerTest', description='Security Manager controller test')
class SecurityManagerTestController:

    @ControllerMethod(
        url = f'/consume',
        responseClass = dict,
        logRequest = True,
        logResponse = True,
        roleRequired=['TEST_ROLE']
    )
    def get(self):
        return {
            'secured': 'information',
            'currentUser': SecurityManager.getCurrentUser()
        }, HttpStatus.OK

    @ControllerMethod(
        url = f'/login',
        requestClass = dict,
        responseClass = dict,
        logRequest = True,
        logResponse = True
    )
    def post(self, dto):
        headers={'some': 'headers'}
        data = {'some': 'data'}
        return {
            'accessToken': SecurityManager.createAccessToken(dto['id'], ['TEST_ROLE'], deltaMinutes=VALID_TOKEN_MINUTES_DURATION, headers=headers, data=data)
        }, HttpStatus.OK

    @ControllerMethod(
        url = f'/refresh',
        requestClass = dict,
        responseClass = dict,
        logRequest = True,
        logResponse = True,
        roleRequired=['TEST_ROLE']
    )
    def patch(self, dto):
        assert 'headers' == SecurityManager.getJwtHeaders().get('some'), f"headers == {SecurityManager.getJwtHeaders().get('some')} should be equals. Headers: {SecurityManager.getJwtHeaders()}"
        headers={'some': 'other headers'}
        data = {'some': 'other data'}
        return {
            'accessToken': SecurityManager.patchAccessToken(newContextList=['TEST_ROLE', 'TEST_ROLE_REFRESH'], headers=headers, data=data)
        }, HttpStatus.OK

    @ControllerMethod(
        url = f'/logout',
        requestClass = dict,
        responseClass = dict,
        logRequest = True,
        logResponse = True,
        roleRequired=['TEST_ROLE', 'TEST_ROLE_REFRESH']
    )
    def put(self, dto):
        SecurityManager.addAccessTokenToBlackList()
        return {'message': 'Logged out'}, HttpStatus.ACCEPTED


@Controller(url=f'/test/{EnvironmentHelper.get("URL_VARIANT")}/security-manager', tag='SecurityManagerTest', description='Security Manager controller test')
class SecurityManagerTestBatchController:

    @ControllerMethod(
        url = f'/consume/only-after-refresh',
        responseClass = dict,
        logRequest = True,
        logResponse = True,
        roleRequired=['TEST_ROLE_REFRESH']
    )
    def get(self):
        assert 'other headers' == SecurityManager.getJwtHeaders().get('some'), f"other headers == {SecurityManager.getJwtHeaders().get('some')} should be equals. Headers: {SecurityManager.getJwtHeaders()}"
        return {
            'secured': 'information',
            'after': 'refresh',
            'currentUser': SecurityManager.getCurrentUser()
        }, HttpStatus.OK



from python_helper import EnvironmentHelper, log

from python_helper import SettingHelper

import globals
globalsInstance = globals.newGlobalsInstance(__file__
    , loadLocalConfig = SettingHelper.activeEnvironmentIsDefault()

    , settingStatus = True
    , statusStatus = True
    , infoStatus = True
    , successStatus = True
    , errorStatus = True
    , warningStatus = True

    , debugStatus = True
    , failureStatus = True
    , wrapperStatus = True
    , logStatus = True
    , testStatus = True
)

from python_framework.api.src.service.flask import FlaskManager
import TestApi
app = TestApi.app

@FlaskManager.initialize(app.api, defaultUrl = '/swagger', openInBrowser=False)
def runFlaskApplication(app=None):
    FlaskManager.runApi(debug=False, use_reloader=False)
    # app.run(debug=False, use_reloader=False)
    # app.run(debug=True)

if __name__ == '__main__' :
    runFlaskApplication()

log.debug(log.debug, f'variant: {EnvironmentHelper.get("URL_VARIANT")}')



import inspect, functools
from python_helper import Constant as c
from python_helper import ReflectionHelper, ObjectHelper, StringHelper, log, Function
from python_framework.api.src.service.ExceptionHandler import GlobalException, DEFAULT_MESSAGE, DEFAULT_LOG_MESSAGE
from python_framework.api.src.enumeration.HttpStatus import HttpStatus


@Function
def EncapsulateItWithGlobalException(message=DEFAULT_MESSAGE, status=HttpStatus.INTERNAL_SERVER_ERROR) :
    encapsulateItWithGlobalExceptionStatus = status
    def wrapper(function,*args,**kwargs) :
        log.wrapper(wrapper, f'Wrapping {function} function with (*{args} args and **{kwargs} kwargs)')
        def wrapedFunction(*args,**kwargs) :
            try :
                functionReturn = function(*args,**kwargs)
            except Exception as exception :
                if isinstance(exception, GlobalException):
                    raise exception
                logMessage = str(exception) if StringHelper.isNotBlank(str(exception)) else LOG_MESSAGE_NOT_PRESENT
                functionName = ReflectionHelper.getName(function, typeName=c.TYPE_FUNCTION)
                log.wrapper(EncapsulateItWithGlobalException, f'''Failed to execute "{functionName}(args={args}, kwargs={kwargs})" {c.TYPE_FUNCTION} call''', exception)
                raise GlobalException(
                    message=message,
                    logMessage=logMessage,
                    logResource=ReflectionHelper.getParentClass(function),
                    logResourceMethod=function,
                    status=HttpStatus.map(encapsulateItWithGlobalExceptionStatus).enumValue
                )
            return functionReturn
        ReflectionHelper.overrideSignatures(wrapedFunction, function)
        return wrapedFunction
    return wrapper



from python_helper import Constant as c
from python_helper import StringHelper

from python_framework.api.src.enumeration.WeekDay import WeekDay


ALL_WEEK = '*'

WEEK_DAY_LIST = [
    WeekDay.MONDAY
    , WeekDay.TUESDAY
    , WeekDay.WEDNESDAY
    , WeekDay.THURSDAY
    , WeekDay.FRIDAY
]

WEEKEND_DAY_LIST = [
    WeekDay.SUNDAY
    , WeekDay.SATURDAY
]

WEEK_LIST = [
    *WEEK_DAY_LIST,
    *WEEKEND_DAY_LIST
]

WEEK_DAY_CHRON = StringHelper.join(
    [weekDay.short for weekDay in WEEK_DAY_LIST],
    character = c.COMA
)

WEEKEND_DAY_CHRON = StringHelper.join(
    [weekDay.short for weekDay in WEEKEND_DAY_LIST],
    character = c.COMA
)

WEEK_CHRON = StringHelper.join(
    [weekDay.short for weekDay in WEEK_LIST],
    character = c.COMA
)



class HeaderKey:
    CONTENT_TYPE = 'Content-Type'

class HeaderValue:
    APPLICATION_JSON = 'application/json'
    MULTIPART_X_MIXED_REPLACE = 'multipart/x-mixed-replace'
    TEXT_HTML = 'text/html; charset=UTF-8'
    TEXT_PLAIN = 'text/plain'
    FROM_DATA = 'multipart/form-data; boundary=value'
    DEFAULT_CONTENT_TYPE = APPLICATION_JSON


class Verb:
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
    PATCH = 'PATCH'
    OPTIONS = 'OPTIONS'

CONTROLLER_CONTEXT = 'Controller'
CLIENT_CONTEXT = 'Client'
LISTENER_CONTEXT = 'Listener'
EMITTER_CONTEXT = 'Emmiter'


REQUEST_HEADERS_KEY = 'requestHeaders'
RESPONSE_HEADERS_KEY = 'responseHeaders'
REQUEST_BODY_KEY = 'requestBody'
RESPONSE_BODY_KEY = 'responseBody'
